/*
 * acsctl - peripheral control for ACR1281U-C1 (ACS ACR1281 1S Dual Reader)
 *
 * (c) 2016 Andreas Steinmetz, ast@domdv.de
 * License: public domain
 *
 * Note: to use this utility you need:
 *
 * 1. the acsccid driver, see http://acsccid.sourceforge.net/
 * 2. have DRIVER_OPTION_CCID_EXCHANGE_AUTHORIZED in Info.plist
 *    This file is typically installed (on a 64 bit system) as:
 *    /usr/lib64/readers/usb/ifd-acsccid.bundle/Contents/Info.plist
 *    The default value for "ifdDriverOptions" which is "0x0000"
 *    needs to be changed to "0x0001".
 *
 * For documentation of the commands, see:
 * "ACR1281U-C1 - Application Programming Interface Version 1.06"
 * http://www.acs.com.hk/
 *
 * No guarantees, whatsoever. If you brick your reader using this code
 * I'm not responsible.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <pcsclite.h>
#include <winscard.h>
#include <reader.h>

#ifndef IOCTL_SMARTCARD_VENDOR_IFD_EXCHANGE
#define IOCTL_SMARTCARD_VENDOR_IFD_EXCHANGE SCARD_CTL_CODE(1)
#endif

typedef struct
{
	SCARDCONTEXT card;
	SCARDHANDLE handle;
	int lock;
} CTX;

typedef struct
{
	char *name;
	int atrlen;
	unsigned char atr[MAX_ATR_SIZE];
} LIST;

static int init_pcsc(void **ctx)
{
	CTX *_ctx;

	if(!ctx||!(_ctx=malloc(sizeof(CTX))))return -1;
	memset(_ctx,0,sizeof(CTX));

	if(SCardEstablishContext(SCARD_SCOPE_SYSTEM,NULL,NULL,&_ctx->card)
		!=SCARD_S_SUCCESS)return -1;
	*ctx=_ctx;
	return 0;
}

static void fini_pcsc(void *ctx)
{       
	CTX *_ctx=ctx;
	
	if(!ctx)return;
	SCardReleaseContext(_ctx->card);
	free(ctx);
}

static int connect_pcsc(void *ctx,char *name)
{       
	CTX *_ctx=ctx;
	DWORD unused;

	if(!ctx||!name)return -1;
	if(SCardConnect(_ctx->card,name,SCARD_SHARE_DIRECT,
		SCARD_PROTOCOL_RAW,&_ctx->handle,&unused)!=
			SCARD_S_SUCCESS)return -1;
	return 0;
}

static void disconnect_pcsc(void *ctx)
{
	CTX *_ctx=ctx;

	SCardDisconnect(_ctx->handle,SCARD_LEAVE_CARD);
}

static int list_pcsc(void *ctx,LIST **readers,int *total)
{
	int i;
	int j;
	char *ptr;
	LIST *arr;
	CTX *_ctx=ctx;
	DWORD len=SCARD_AUTOALLOCATE;
	LPSTR *data;
	SCARD_READERSTATE state;

	if(!ctx||!readers)return -1;
	if(SCardListReaders(_ctx->card,NULL,(LPSTR)&data,&len)!=
		SCARD_S_SUCCESS)return -1;
	for(*total=0,ptr=(char *)data,i=0;i<len;i++)if(!ptr[i])(*total)++;
	if(!*total||!--(*total))
	{
		*readers=NULL;
		return 0;
	}
	if(!(*readers=malloc(*total*sizeof(LIST)+len)))
	{
		SCardFreeMemory(_ctx->handle,data);
		return -1;
	}
	ptr=((char *)(*readers))+*total*sizeof(LIST);
	memcpy(ptr,data,len);
	SCardFreeMemory(_ctx->handle,data);
	arr=*readers;
	arr[0].name=ptr;
	for(j=1,i=1;i<len&&j<*total;i++)if(!ptr[i])arr[j++].name=ptr+i+1;
	for(j=0;j<*total;j++)
	{
		arr[j].atrlen=0;
		state.szReader=arr[j].name;
		state.dwCurrentState=SCARD_STATE_UNAWARE;
		if(SCardGetStatusChange(_ctx->card,INFINITE,&state,1)!=
			SCARD_S_SUCCESS)continue;
		if(((state.dwEventState&0xffff)&
		   (SCARD_STATE_PRESENT|SCARD_STATE_EXCLUSIVE|SCARD_STATE_MUTE))
			!=SCARD_STATE_PRESENT)continue;
		if(!state.cbAtr)continue;
		arr[j].atrlen=state.cbAtr;
		memcpy(arr[j].atr,state.rgbAtr,arr[j].atrlen);
	}
	return 0;
}


static int pcsc_ctrl(void *ctx,unsigned char *in,int ilen,unsigned char *out,
	int *olen)
{
	CTX *_ctx=ctx;
	DWORD len;

	len=*olen;

	if(SCardControl(_ctx->handle,IOCTL_SMARTCARD_VENDOR_IFD_EXCHANGE,
		in,ilen,out,len,&len)!=SCARD_S_SUCCESS)return -1;
	*olen=len;
	return 0;
}

static int print_fw(void *ctx)
{
	unsigned char getfw[5]={0xe0,0x00,0x00,0x18,0x00};
	unsigned char fwans[4]={0xe1,0x00,0x00,0x00};
	unsigned char reply[256];
	int len=sizeof(reply)-1;

	if((pcsc_ctrl(ctx,getfw,5,reply,&len)))return -1;
	if(len<5)return -1;
	if(memcmp(reply,fwans,4))return -1;
	if(len!=reply[4]+5)return -1;
	reply[len]=0;
	printf("%s\n",reply+5);
	return 0;
}

static int led_ctrl(void *ctx,int state)
{
	unsigned char ledctl[6]={0xe0,0x00,0x00,0x29,0x00};
	unsigned char ledans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(state<0||state>255)return -1;
	ledctl[5]=(unsigned char)state;
	if((pcsc_ctrl(ctx,ledctl,6,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,ledans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int led_status(void *ctx)
{
	unsigned char ledstat[5]={0xe0,0x00,0x00,0x29,0x00};
	unsigned char ledans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if((pcsc_ctrl(ctx,ledstat,5,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,ledans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int buzzer_ctrl(void *ctx,int duration)
{
	unsigned char buzzerctl[6]={0xe0,0x00,0x00,0x28,0x01};
	unsigned char buzzerans[6]={0xe1,0x00,0x00,0x00,0x01,0x00};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(duration<1||duration>255)return -1;
	buzzerctl[5]=(unsigned char)duration;
	if((pcsc_ctrl(ctx,buzzerctl,6,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,buzzerans,6))return -1;
	if(len!=reply[4]+5)return -1;
	return 0;
}

static int ledbuzz_behavior(void *ctx)
{
	unsigned char ledbuzzbhv[5]={0xe0,0x00,0x00,0x21,0x00};
	unsigned char ledbuzzans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if((pcsc_ctrl(ctx,ledbuzzbhv,5,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,ledbuzzans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int ledbuzz_default(void *ctx,int mode)
{
	unsigned char ledbuzzbhv[6]={0xe0,0x00,0x00,0x21,0x01};
	unsigned char ledbuzzans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(mode<0||mode>255)return -1;
	ledbuzzbhv[5]=(unsigned char)mode;
	if((pcsc_ctrl(ctx,ledbuzzbhv,6,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,ledbuzzans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int set_inscnt(void *ctx,int icc,int picc)
{
	unsigned char insctl[9]={0xe0,0x00,0x00,0x09,0x04};
	unsigned char insans[5]={0xe1,0x00,0x00,0x00,0x00};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(icc<0||icc>65535||picc<0||picc>65535)return -1;
	insctl[5]=(unsigned char)(icc&0xff);
	insctl[6]=(unsigned char)(icc>>8);
	insctl[7]=(unsigned char)(picc&0xff);
	insctl[8]=(unsigned char)(picc>>8);
	if((pcsc_ctrl(ctx,insctl,9,reply,&len)))return -1;
	if(len!=5)return -1;
	if(memcmp(reply,insans,5))return -1;
	if(len!=reply[4]+5)return -1;
	return 0;
}

static int get_inscnt(void *ctx)
{
	unsigned char insreq[5]={0xe0,0x00,0x00,0x09,0x00};
	unsigned char insans[5]={0xe1,0x00,0x00,0x00,0x04};
	unsigned char reply[16];
	int len=sizeof(reply);
	int icc;
	int picc;

	if((pcsc_ctrl(ctx,insreq,5,reply,&len)))return -1;
	if(len!=9)return -1;
	if(memcmp(reply,insans,5))return -1;
	if(len!=reply[4]+5)return -1;
	icc=reply[6];
	icc<<=8;
	icc|=reply[5];
	picc=reply[8];
	picc<<=8;
	picc|=reply[7];
	printf("%04x %04x\n",icc,picc);
	return 0;
}

static int upd_inscnt(void *ctx)
{
	unsigned char insupd[5]={0xe0,0x00,0x00,0x0a,0x00};
	unsigned char insans[5]={0xe1,0x00,0x00,0x00,0x04};
	unsigned char reply[16];
	int len=sizeof(reply);
	int icc;
	int picc;

	if((pcsc_ctrl(ctx,insupd,5,reply,&len)))return -1;
	if(len!=9)return -1;
	if(memcmp(reply,insans,5))return -1;
	if(len!=reply[4]+5)return -1;
	icc=reply[6];
	icc<<=8;
	icc|=reply[5];
	picc=reply[8];
	picc<<=8;
	picc|=reply[7];
	printf("%04x %04x\n",icc,picc);
	return 0;
}

static int set_autopoll(void *ctx,int mode)
{
	unsigned char autoset[6]={0xe0,0x00,0x00,0x23,0x01};
	unsigned char autoans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(mode<0||mode>255)return -1;
	autoset[5]=(unsigned char)mode;
	if((pcsc_ctrl(ctx,autoset,6,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,autoans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int read_autopoll(void *ctx)
{
	unsigned char autoget[5]={0xe0,0x00,0x00,0x23,0x00};
	unsigned char autoans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if((pcsc_ctrl(ctx,autoget,5,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,autoans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int manual_poll(void *ctx)
{
	unsigned char manupoll[6]={0xe0,0x00,0x00,0x22,0x01,0x0a};
	unsigned char pollans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if((pcsc_ctrl(ctx,manupoll,6,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,pollans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int set_piccop(void *ctx,int mode)
{
	unsigned char piccset[6]={0xe0,0x00,0x00,0x20,0x01};
	unsigned char piccans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(mode<0||mode>255)return -1;
	piccset[5]=(unsigned char)mode;
	if((pcsc_ctrl(ctx,piccset,6,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,piccans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int read_piccop(void *ctx)
{
	unsigned char piccget[5]={0xe0,0x00,0x00,0x20,0x00};
	unsigned char piccans[5]={0xe1,0x00,0x00,0x00,0x01};
	unsigned char reply[16];
	int len=sizeof(reply);

	if((pcsc_ctrl(ctx,piccget,5,reply,&len)))return -1;
	if(len!=6)return -1;
	if(memcmp(reply,piccans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x\n",reply[5]);
	return 0;
}

static int set_exclusive(void *ctx,int mode)
{
	unsigned char exclset[6]={0xe0,0x00,0x00,0x2b,0x01};
	unsigned char exclans[5]={0xe1,0x00,0x00,0x00,0x02};
	unsigned char reply[16];
	int len=sizeof(reply);

	if(mode<0||mode>1)return -1;
	exclset[5]=(unsigned char)mode;
	if((pcsc_ctrl(ctx,exclset,6,reply,&len)))return -1;
	if(len!=7)return -1;
	if(memcmp(reply,exclans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x %02x\n",reply[5],reply[6]);
	return 0;
}

static int read_exclusive(void *ctx)
{
	unsigned char exclget[5]={0xe0,0x00,0x00,0x2b,0x00};
	unsigned char exclans[5]={0xe1,0x00,0x00,0x00,0x02};
	unsigned char reply[16];
	int len=sizeof(reply);

	if((pcsc_ctrl(ctx,exclget,5,reply,&len)))return -1;
	if(len!=7)return -1;
	if(memcmp(reply,exclans,5))return -1;
	if(len!=reply[4]+5)return -1;
	printf("%02x %02x\n",reply[5],reply[6]);
	return 0;
}

static struct
{
	char *cmd;
	int parms;
	int p816;
	int (*func)();
} cmds[]=
{
	{"get-firmware-version",   0,0,print_fw        },
	{"led-control",            1,0,led_ctrl        },
	{"led-status",             0,0,led_status      },
	{"buzzer-control",         1,0,buzzer_ctrl     },
	{"set-led-buzzer-default", 1,0,ledbuzz_default },
	{"read-led-buzzer-default",0,0,ledbuzz_behavior},
	{"initialize-counters",    2,1,set_inscnt      },
	{"read-counters",          0,0,get_inscnt      },
	{"update-counters",        0,0,upd_inscnt      },
	{"set-auto-picc-polling",  1,0,set_autopoll    },
	{"read-auto-picc-polling", 0,0,read_autopoll   },
	{"manual-pic-polling",     0,0,manual_poll     },
	{"set-picc-parameters",    1,0,set_piccop      },
	{"read-picc-parameters",   0,0,read_piccop     },
	{"set-exclusive-mode",     1,0,set_exclusive   },
	{"read-exclusive-mode",    0,0,read_exclusive  },
	{NULL,                     0,0,NULL            }
};

int main(int argc,char*argv[])
{
	void *ctx;
	LIST *list;
	int total;
	int i;
	int r=0;
	int p1=0;
	int p2=0;
	unsigned long int val;
	char *end;
	char *reader="ACS ACR1281 1S Dual Reader 00 00";

	if(init_pcsc(&ctx))
	{
		fprintf(stderr,"pcsc init failure\n");
		return 1;
	}

	if(argc>2&&!strcmp(argv[1],"-r"))
	{
		reader=argv[2];
		argc+=2;
		argv-=2;
	}

	if(argc<2)
	{
		fprintf(stderr,"no parameters, see source for help\n");
		fini_pcsc(ctx);
		return 1;
	}

	if(argc==2&&!strcmp(argv[1],"list-readers"))
	{
		if(list_pcsc(ctx,&list,&total))
		{
			fprintf(stderr,"pcsc reader list failed\n");
			fini_pcsc(ctx);
			return 1;
		}

		for(i=0;i<total;i++)printf("%s\n",list[i].name);
		free(list);
		fini_pcsc(ctx);
		return 0;
	}

	for(i=0;cmds[i].cmd;i++)if(!strcmp(argv[1],cmds[i].cmd))break;

	if(!cmds[i].cmd)
	{
		fprintf(stderr,"unknown command\n");
		fini_pcsc(ctx);
		return 1;
	}

	if(argc!=cmds[i].parms+2)
	{
		fprintf(stderr,"wrong parameter count for command\n");
		fini_pcsc(ctx);
		return 1;
	}

	switch(cmds[i].parms)
	{
	case 2:	val=strtoul(argv[3],&end,0);
		if(end==argv[3]||*end||val>(cmds[i].p816?65535:255))
		{
			fprintf(stderr,"parameter value error\n");
			fini_pcsc(ctx);
			return 1;
		}
		p2=(int)val;

	case 1:	val=strtoul(argv[2],&end,0);
		if(end==argv[2]||*end||val>(cmds[i].p816?65535:255))
		{
			fprintf(stderr,"parameter value error\n");
			fini_pcsc(ctx);
			return 1;
		}
		p1=(int)val;
	}

	if(connect_pcsc(ctx,reader))
	{
		fprintf(stderr,"connect to reader failed\n");
		fini_pcsc(ctx);
		return 1;
	}

	switch(cmds[i].parms)
	{
	case 0:	r=cmds[i].func(ctx);
		break;

	case 1:	r=cmds[i].func(ctx,p1);
		break;

	case 2:	r=cmds[i].func(ctx,p1,p2);
		break;
	}

	if(r)
	{
		fprintf(stderr,"reader command failed\n");
		disconnect_pcsc(ctx);
		fini_pcsc(ctx);
		return 1;
	}

	disconnect_pcsc(ctx);
	fini_pcsc(ctx);
	return 0;
}
